#pragma once

// We are using the allegro library for graphics
#define BUILD_ALLEGRO
//#define BUILD_OPENGL

#include "GFX.h"
#include "Enums.h"
#include "Math.h"
#include "Vec2.h"
#include <list>

// Our constants (Resolution among them for temporarily)
#define FPS 60
#define SC_WIDTH 1280
#define SC_HEIGHT 1024
#define CELL_SIZE 16
#define CELLS_X SC_WIDTH / CELL_SIZE
#define CELLS_Y SC_HEIGHT / CELL_SIZE

struct Color;
class Entity;
class Renderer;
class Particle;
typedef std::list<Entity *> Entities;
typedef std::list<Renderer *> Renderers;

// Our game class, which is the main manager of all things game-related.
class Game {
public:
	static Game* Instance();

	// Public vars
	GameState state;

	// Helper Functions
	static Vec2 DirToVec2(Direction dir);
	Vec2 Game::GetCenterOfCell(Vec2 pos);
	Entity* GetGridAt(Vec2 pos);
	int GetTerrainAt(Vec2 pos);
	bool InGrid(Vec2 pos);
	Color GetColorOfTeam(Team team);
	std::string GetTeamName(Team team);
	bool ScreenPointInGrid(int px, int py, int gx, int gy);
	Vec2 ScreenPointToGrid(int px, int py);

	// Game Functions
	void OrderTeam(Team team, Vec2 pos);
	void EntityMoved(Vec2 from, Vec2 to, Entity* ent);
	Vec2 GetClosestFriendlyTarget(Entity *me, int dist);
	void TeamEliminated(Team team);


	// Logic/Render
	void Update();
	void Render();

	// Renderer Management
	Entity* AddEntity(Entity *newEntity);
	void DeleteObject(Renderer *renderer, bool safe = false);
	void CreateObject(Renderer *renderer);
	void CreateExplosion(Vec2 pos);

protected:
	Game();
private:
	// Object lists
	Renderers objects;
	Entities entities;
	Renderers deletionQueue;

	// Private vars
	Team loser;
	Color gridcol;

	// Rendering Functions
	void drawGrid();
	void drawTerrain();
	void drawHUD();

	// Intenal Management Functions
	void CleanUp();
	void UpdateTerrain();
	void DestroyEntity(Entity *entity);
	void DestroyParticle(Particle *particle);

	// Our game field
	Entity* grid[CELLS_X][CELLS_Y];
	int terrain[CELLS_X][CELLS_Y];

	// Singleton instance
	static Game* instance;
};

// The (abstract) base class for all objects which draw to the screen.
class Renderer {
public:
	int x;
	int y;
	Color col;

	Renderer() { };
	virtual ~Renderer(){ };
	virtual void Render() = 0;
	virtual void Update() = 0;
	virtual void Destroy() = 0;
	virtual RenderType GetType() = 0;
};

class Particle;

typedef std::list<Particle *> Particles;

// Particle class, used for objects which are not on the grid (particle effects/projectiles)
class Particle : public Renderer {
public:
	int time;
	int fade;
	int size;
	Vec2 pos;
	Vec2 force;
	Vec2 dir; // normalized version of force
	ParticleType type;

	void Render();
	void Update();
	RenderType GetType() { return RT_PARTICLE; };
	Particle();
	void Destroy();
	Particle(int size, int time, Color color, int forcelow, int forcehigh, Vec2 pos, int fade, ParticleType type=PT_SQUARE);
};

// Projectiles are like particles, but with a target and a source.
class Projectile : public Particle {
public:
	Team team;
	float dmg;
	Entity* from;
	Projectile();
	Projectile(Entity *from, float dmg, int size, Color color, int force, Vec2 pos, Vec2 target, ParticleType type=PT_SQUARE);
	void Update();
private:
	Vec2 target;
};

// TODO: Homing missiles? Might not be necessary at this point, since units tend to not move around too much
// ..although it would look cool
class Missile : public Projectile {
public:
	Missile(Entity* from, float dmg, int size, Color color, int force, Vec2 pos, Vec2 target, ParticleType type);
};

// TODO: Create [newParticle] every [time] frames, [amount] times 
// Can just do update calls to create new particles, don't do render call, and let particles be managed
// by the game.
class ParticleSpawner : public Renderer {
public:
	ParticleSpawner(Particle newParticle, int time, int amount);
private:
	Projectile spawn;
};

extern GFX *gfx;