/* Particles.cpp
	Contains info on the particle class and its derivatives. Particles do not exist on the grid. They also manage
	themselves in most ways *EXCEPT* for creation and deletion which is managed by the objects list in Game.
*/

#include "Game.h"
#include "Entities.h"
#include <Math.h>

// Default Particle Constructor
Particle::Particle() {
	this->size = 1;
	this->time = 0;
	this->col = Color(0,0,0);
	this->force = Vec2(0,0);
	this->pos = Vec2(0,0);
	this->fade = 0;
	this->type = PT_SQUARE;
	this->dir = Vec2(0,0);
}

// Main Particle Constructor
Particle::Particle(int size, int time, Color color, int forcelow, int forcehigh, Vec2 pos, int fade, ParticleType type) {
	this->size = size;
	this->time = time;
	this->col = color;
	int xmod = (rand() % 2 == 1) ? -1 : 1;
	int ymod = (rand() % 2 == 1) ? -1 : 1;
	this->force = Vec2((forcelow +  rand() % forcehigh) * xmod, (forcelow +  rand() % forcehigh) * ymod);
	this->pos = pos;
	this->fade = fade;
	this->type = type;
}

// Particle update. Changes position by [force], and darkens color by [fade]
// If time <= 0, tells game to delete it.
void Particle::Update() {
	this->pos.x += force.x;
	this->pos.y += force.y;
	col = col - Color(fade, fade, fade);

	time--;
	if(time <= 0) {		
		Game::Instance()->DeleteObject(this);
	}
}

void Particle::Destroy() {
	
}

// Handles the rendering. We can do square particles or lines. (for now)
void Particle::Render() {
	switch(type) {
		case PT_SQUARE:
			if(size == 1)
				gfx->GFXputpixel(pos.x, pos.y, col);
			else
				gfx->GFXrectfill(pos.x - size/2, pos.y - size/2, pos.x + size/2, pos.y + size/2, col);
		break;
		case PT_LINE:
			//if(dir == Vec2(0,0))
				gfx->GFXline(pos.x, pos.y, pos.x + force.x, pos.y + force.y, col);
			//else {
			//	Vec2 lineLen = dir * size;
			//	gfx->GFXline(pos.x, pos.y, pos.x + lineLen.x, pos.y + lineLen.y, col);
			//}
		break;
	}
}

// Projectiles are similar to particles, however, they also have information about which entity they have come from
// and a target which they head to.
Projectile::Projectile(Entity* from, float dmg, int size, Color color, int force, Vec2 pos, Vec2 target, ParticleType type)
: Particle(size, 0, color, 2, 2, pos, 0, type){
	
	this->from = from;
	this->dmg = dmg;
	if(from)
		this->team = from->team;
	
	// Find out our projectile's direction
	Vec2 direction = target - pos;

	// Make it a unit vector so we can scale it
	direction = direction.Normalize();
	
	// Save the unit vector
	dir = direction;

	// Then, calculate our force (velocity) vector, as well as approximating the time the projectile should stay alive
	// as we don't want entities who are able to hit stuff way out of their range.
	this->force = direction * force;
	float dist = sqrt(target.distSqr(pos));
	this->time = (int)(dist / force) + force;
	//this->fade = (int)(128 / time);
}

// Default Constructor
Projectile::Projectile() : Particle() {
	this->from = 0;
	this->dmg = 0;
	this->team = TEAM_NONE;
}

// Call parent's update, then check the current grid node this projectile is in. If we're on an
// enemy, deal damage and queue for deletion!
void Projectile::Update() {
	Particle::Update();

	Entity *ent = Game::Instance()->GetGridAt(Game::Instance()->ScreenPointToGrid(pos.x, pos.y));
	if(ent && ent->team != team) {
		ent->TakeDamage(from, dmg);
		Game::Instance()->DeleteObject(this, true);
	}
}

// Homing missiles? Not implemented.
Missile::Missile(Entity* from, float dmg, int size, Color color, int force, Vec2 pos, Vec2 target, ParticleType type) : Projectile(){
	this->from = from;
	this->dmg = dmg;
	this->size = size;
	this->col = color;

}